/**
FAT Filesystem Traversal & Data Extraction routines

Orion Sky Lawlor, olawlor@acm.org, 2006/04/02 (Public Domain)
*/
#define _FILE_OFFSET_BITS 64 /* 64-bit file access in Linux */
#define _LARGEFILE_SOURCE 1 /* ditto... */
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <algorithm> /* for std::min and std::max */
#include <string>
#include <vector>
#include "fat_layout.h"
#include "osl/mkdir.h" /* for making directories */

#undef min /* friggin' Windows... */
#undef max

/* If set to 1, prints out some debugging information. 
  at level 2, prints DOS directory entries too.
  at level 3, prints windows long filename directory entries too.
  at level 4, prints out every disk access
*/
int verbose=1;

/* If set to 1, saves file contents to out/ directory. */
int dump_files=1;




/** A FAT filesystem image, stored on disk. */
class diskimage_fat : public fat_filesystem {
	FILE *f;
	fileoff_t start_jump;
public:
	diskimage_fat(FILE *f_,fileoff_t jump)  :f(f_),start_jump(jump) 
	{
		if (f==NULL) abort("unable to open disk image");
	}
	virtual void setup(void) {
		fat_filesystem::setup();
		if (verbose) {
			std::cout<<"FAT filesystem layout:\n";
			std::cout<<"  FAT consists of "<<fat_count<<" entries\n";
			std::cout<<"  Each entry is for one cluster of "<<cluster_size<<" bytes\n";
			std::cout<<"  Clusters start on disk at byte "<<cluster_start<<"\n";
			std::cout<<"  First few FAT entries: \n";
			for (int i=0;i<20;i++) std::printf("      fat[%X]=%X \n",i,(int)fat[i]);
			std::cout<<"\n\n";
		}
	}
	virtual void read_disk(void *p,unsigned int n,fileoff_t o) {
		if (n==0) return;
		if (verbose>=4) std::cout<<"       read_disk("<<p<<","<<n<<","<<o<<");\n";
		// FIXME: where possible, use fseeko(f,o,SEEK_SET);
		fseek(f,o+start_jump,SEEK_SET);
		int read=fread((char *)p,1,n,f);
		if (read<=0) {
			std::cout<<"Read failure reading "<<n<<" bytes at offset "<<o<<"\n";
			abort("unable to read disk");
		}
	}
};


fat_filesystem::fat_filesystem() {fat=NULL;root_dir=NULL;}
fat_filesystem::~fat_filesystem() {delete[] fat;delete root_dir;}

/* Read the boot sector and FAT */
void fat_filesystem::setup(void) {
	/* The boot sector is the first thing in the disk */
	if (verbose) std::cout<<"Loading boot sector...\n";
	read_disk(&boot_sector,sizeof(boot_sector),0);
	if (verbose) std::cout<<"Boot sector by '"<<boot_sector.system_id<<"'\n";
	if (boot_sector.system_id[0]<=' ')
		std::cout<<"WARNING: Bogus boot sector system id!  Is this a FAT filesystem image? Do you need '-p'?\n";
	
	/* Figure out the sizes of various pieces.
	  FIXME: do a sanity check for FAT32 somehow?
	*/
	sector_size=boot_sector.sector_size;
	cluster_size=boot_sector.sec_per_clus*sector_size;
	fileoff_t fat_start=boot_sector.reserved*sector_size;
	fileoff_t fat_size=boot_sector.fat32_length*sector_size;
	fat_count=fat_size/sizeof(fat_t);
	cluster_start=fat_start+fat_size*boot_sector.fats;
	int root_cluster=boot_sector.root_cluster;
	cluster_start-=root_cluster*cluster_size; /* silly-- first cluster on disk is "root_cluster", not 0 */
	
	/* Read the FAT */
	fat=new fat_t[fat_count];
	read_disk(&fat[0],fat_size,fat_start);
	
	/* Read subsequent FATs and check their entries */
	for (int fatNo=1;fatNo<boot_sector.fats;fatNo++) {
		if (verbose) std::cout<<"Comparing primary to FAT #"<<fatNo+1<<"\n";
		fat_t *f2=new fat_t[fat_count];
		read_disk(&f2[0],fat_size,fat_start+fat_size);
		for (int i=0;i<fat_count;i++)
			if (fat[i]!=f2[i]) 
				abort("First and second FAT entries don't match!\n");
		delete[] f2;
	}
	
	/* The root directory entry starts at the first disk cluster */
	root_dir=new fat_file(this,root_cluster,~0);
}

/** Something went horribly wrong-- leave. */
void fat_filesystem::abort(const char *why) {
	std::cout<<"Fatal error:"<<why<<"\n";
	std::exit(1);
}

/******************** Files ***************/

/** FAT file constructor */
fat_file::fat_file(fat_filesystem *fs_,cluster_t first,fileoff_t length)
	:fs(fs_), cluster_size(fs->get_cluster_size()),
	 clus(first), clus_left(cluster_size), file_left(length) {}

/** Read up to n bytes from the file into p
*/
unsigned int fat_file::read_partial(void *p,unsigned int wanted)
{
	char *start=(char *)p;
	char *dest=start;
	while (1) { /* Keep reading clusters of data from file */
		/* Figure out how many bytes we can read, n */
		fileoff_t n=std::min(std::min(wanted,(unsigned int)file_left),clus_left); 
		fs->read_disk(dest,n,
			fs->cluster_location(clus)+cluster_size-clus_left);
		/* Advance to the next set of bytes */
		dest+=n;
		wanted-=n;
		file_left-=n;
		clus_left-=n;
		if (clus_left==0) { /* this cluster's empty--advance */
			clus=fs->fat_advance(clus); /* go to next cluster */
			if (clus) clus_left=cluster_size;
		}
		if (wanted==0) return dest-start; /* Got everything we wanted */
		if (file_left==0) return dest-start; /* Got everything in the file */
		if (clus==0) return dest-start; /* Fell off end of last cluster */
	}
}


/********************* Directories **********************/

/* Convert UNICODE little-endian 16-bit characters to std::string */
std::string winName(const byte *data,int nChar) {
	std::string ret;
	for (int i=0;i<nChar;i++) {
		char c=(char)data[2*i+0]; /* low byte of 16-bit unicode */
		if (c==(char)0 || c==(char)0xff) continue;
		ret+=c;
	}
	return ret;
}
std::string winName(win_dir_entry &e) {
	return winName(e.name0_4,5)+winName(e.name5_10,6)+winName(e.name11_12,2);
}
std::string dosName(msdos_dir_entry &e) {
	std::string ret;
	for (int i=0;i<8;i++) 
		if (e.name[i]>' ') /* names are space-padded */
			ret+=(char)e.name[i];
	ret+='.';
	for (int j=0;j<3;j++) 
		if (e.ext[j]>' ') 
			ret+=(char)e.ext[j];
	return ret;
}

/** Debugging printout of directory entry */
void printDosDir(msdos_dir_entry &e) {
	std::cout<<"  DOS: name='"<<dosName(e)<<"'  ";
	
	std::cout<<"  attr="<<(int)e.attr<<", "
	    "cluster="<<e.start_cluster()<<","
	    "size="<<(int)e.size<<"\n";
}
void printWinDir(win_dir_entry &e) {
	std::cout<<"  WIN: name='"<<winName(e)
		<<"'  attr="<<(int)e.attr<<"  id="<<(int)e.id<<"\n";
}

/* Default directory consumer does nothing (just ignores everything) */
dir_consumer::~dir_consumer() {}
void dir_consumer::hit_file(fat_file &f,const std::string &name) {}
void dir_consumer::hit_dir(fat_file &f,const std::string &name) {}

/** List everything in this directory, and pass files and directories to dest */
void walk_dir(fat_file &dir,dir_consumer &dest) {
	msdos_dir_entry e;
	std::string name="";
	while (dir.read(&e,sizeof(e))) {
		if (e.attr==ATTR_WIN) { /* volume label bit set-- windows long filename */
			win_dir_entry *w=(win_dir_entry *)&e;
			if (w->id>=65) name=""; /* start of new name */
			if (verbose>=3) printWinDir(*w);
			name=winName(*w)+name;
		} else { /* normal DOS directory entry */
			if (name.size()==0) name=dosName(e); /* no long name*/
			if (name=="." || name=="..") continue; /* dir. link */
			std::cout<<"Long name:'"<<name<<"' ";
			if (verbose>=2) printDosDir(e);
			
			if (e.name[0]==(byte)0 || e.name[0]==(byte)DELETED_FLAG) {
			   /* Deleted file-- skip */
			}
			else if (e.attr&ATTR_DIR) 
			{ /* a directory.  Directories have a bogus zero size */
				fat_file f(dir.get_filesystem(),
					e.start_cluster(),~0);
				dest.hit_dir(f,name);
			}
			else { /* a normal file */
				fat_file f(dir.get_filesystem(),
					e.start_cluster(),e.size);
				dest.hit_file(f,name);
			}
			name="";
		}
	}
}


/** Dump directories and files out of FAT filesystem onto host machine */
class dir_dumper : public dir_consumer {
	std::string path; /* directories encountered so far */
public:
	dir_dumper(std::string initial_path) :path(initial_path) 
	{
		osl::mkdir(path.c_str());
	}
	virtual void hit_file(fat_file &f,const std::string &name)
	{
		std::cout<<"Hit file '"<<path+name<<"'  size "<<f.bytes_left()<<" bytes\n";
		FILE *out=fopen((path+name).c_str(),"wb");
		/* Copy file contents one buf at a time */
		enum {len=4096};
		byte buf[len];
		unsigned int n;
		while (0!=(n=f.read_partial(buf,len)))
			fwrite(buf,1,n,out);
		fclose(out);
	}
	virtual void hit_dir(fat_file &f,const std::string &name)
	{
		std::cout<<"Entering subdirectory "<<name<<"\n";
		std::string old_path=path;
		path=path+name;
		osl::mkdir(path.c_str());
		path=path+"/";
		
		walk_dir(f,*this); /* recursively walk subdirectory files & dirs */
		
		path=old_path;
		std::cout<<"Leaving subdirectory "<<name<<"\n";
	}
};


/***************** Caller code ******************/

void usage(const char *why) {
	std::cout<<"Usage: fat [ -p ] [ -v <level> ] <FAT32 filesystem image>\n"; 
	std::cout<<"   -p: read a single partition image instead of a disk image\n";
	std::cout<<"   -v <level>: set verbosity to <level>\n";
	std::cout<<why<<"\n";
	std::exit(1);
}

int main(int argc,char *argv[]) {
	int argi=1;
	int jump=63*512; /* initial jump to skip partition table */
	if (argc<=argi) usage("Not enough command-line arguments");
	if (0==strcmp(argv[argi],"-p")) 
		{argi++; jump=0;}
	if (0==strncmp(argv[argi],"-v",2)) 
		{argi++; verbose=atoi(argv[argi++]);}
	const char *diskImageName=argv[argi++];
	if (argc<argi) usage("Not enough command-line arguments");
	FILE *diskImage=fopen(diskImageName,"rb");
	diskimage_fat fs(diskImage,jump);
	fs.setup();
	
	/* Dump out contents of root directory */
	std::cout<<"Contents of root directory:\n";
	dir_dumper dumper("out/");
	walk_dir(*fs.root_dir,dumper);
	
	return 0;
}
